@@ -29,20 +29,25 @@
 #   b) syncing with simplenote
 
 # to check if we're online
-
+import sys
 import codecs
-import ConfigParser
+from configparser import ConfigParser
+from .p3port import unicode
 import logging
 from logging.handlers import RotatingFileHandler
-from notes_db import NotesDB, SyncError, ReadError, WriteError
+from .notes_db import NotesDB, SyncError, ReadError, WriteError
 import argparse
 import os
-import sys
-import time
-
-from utils import KeyValueObject, SubjectMixin
-import view
+import traceback
+import threading
+import re
+import typing
+from . import tk
+from .utils import SubjectMixin
+from . import view
 import webbrowser
+from .version import VERSION
+from . import events
 
 try:
     import markdown
@@ -50,6 +55,15 @@ except ImportError:
     HAVE_MARKDOWN = False
 else:
     HAVE_MARKDOWN = True
+    DEFAULT_MARKDOWN_EXTS = (
+        # Add 'fenced code block' syntax support.
+        # If you try to convert without this extension, code block is treated as inline code.
+        # https://python-markdown.github.io/extensions/fenced_code_blocks/
+        'markdown.extensions.fenced_code',
+        # Add table syntax support.
+        # https://python-markdown.github.io/extensions/tables/
+        'markdown.extensions.tables',
+    )
 
 try:
     import docutils
@@ -59,7 +73,22 @@ except ImportError:
 else:
     HAVE_DOCUTILS = True
 
-VERSION = "0.9.7"
+
+class ColorConfig(typing.NamedTuple):
+    # Text color.
+    text: str
+    # Background color of selected note.
+    selected_note: str
+    # Text color for note info (title, tags, updated date, etc).
+    note_info: str
+    # Text color for highlighted note info.
+    highlight_note_info: str
+    # Text color for URL.
+    url: str
+    # Background color.
+    background: str
+    # Background color of highlighted area.
+    highlight_background: str
 
 
 class Config:
@@ -67,6 +96,7 @@ class Config:
     @ivar files_read: list of config files that were parsed.
     @ivar ok: True if config files had a default section, False otherwise.
     """
+
     def __init__(self, app_dir):
         """
         @param app_dir: the directory containing nvpy.py
@@ -76,53 +106,68 @@ class Config:
         # cross-platform way of getting home dir!
         # http://stackoverflow.com/a/4028943/532513
         home = os.path.abspath(os.path.expanduser('~'))
-        defaults = {'app_dir': app_dir,
-                    'appdir': app_dir,
-                    'home': home,
-                    'notes_as_txt': '0',
-                    'housekeeping_interval': '2',
-                    'search_mode': 'gstyle',
-                    'case_sensitive': '1',
-                    'search_tags': '1',
-                    'sort_mode': '1',
-                    'pinned_ontop': '1',
-                    'db_path': os.path.join(home, '.nvpy'),
-                    'txt_path': os.path.join(home, '.nvpy/notes'),
-                    'theme': 'default',
-                    'font_family': 'Courier',  # monospaced on all platforms
-                    'font_size': '10',
-                    'list_font_family': 'Helvetica',  # sans on all platforms
-                    'list_font_family_fixed': 'Courier',  # monospace on all platforms
-                    'list_font_size': '10',
-                    'layout': 'horizontal',
-                    'print_columns': '0',
-                    'background_color': 'white',
-                    'sn_username': '',
-                    'sn_password': '',
-                    'simplenote_sync': '1',
-                    'debug': '1',
-                    # Filename or filepath to a css file used style the rendered
-                    # output; e.g. nvpy.css or /path/to/my.css
-                    'rest_css_path': None,
-                   }
+        defaults = {
+            'app_dir': app_dir,
+            'appdir': app_dir,
+            'home': home,
+            'notes_as_txt': '0',
+            'read_txt_extensions': 'txt,mkdn,md,mdown,markdown',
+            'housekeeping_interval': '2',
+            'search_mode': 'gstyle',
+            'case_sensitive': '1',
+            'search_tags': '1',
+            'sort_mode': '1',
+            'pinned_ontop': '1',
+            'db_path': os.path.join(home, '.nvpy'),
+            'txt_path': os.path.join(home, '.nvpy/notes'),
+            'theme': 'default',
+            'font_family': 'Courier',  # monospaced on all platforms
+            'font_size': '10',
+            'list_font_family': 'Helvetica',  # sans on all platforms
+            'list_font_family_fixed': 'Courier',  # monospace on all platforms
+            'list_font_size': '10',
+            'layout': 'horizontal',
+            'print_columns': '0',
+            'text_color': 'black',
+            'selected_note_color': 'light blue',
+            'note_info_color': 'dark gray',
+            'highlight_note_info_color': 'lightyellow',
+            'url_color': '#03f',
+            'background_color': 'white',
+            'highlight_background_color': 'yellow',
+            'sn_username': '',
+            'sn_password': '',
+            'simplenote_sync': '1',
+            'debug': '1',
+            # Filename or filepath to a css file used style the rendered
+            # output; e.g. nvpy.css or /path/to/my.css
+            'rest_css_path': '',
+            'md_css_path': '',
+            'md_extensions': '',
+            'keep_search_keyword': 'false',
+            'confirm_delete': 'true',
+            'confirm_exit': 'false',
+        }
 
         # parse command-line arguments
         args = self.parse_cmd_line_opts()
 
         # later config files overwrite earlier files
         # try a number of alternatives
-        cfg_files = [os.path.join(app_dir, 'nvpy.cfg'),
-                     os.path.join(home, 'nvpy.cfg'),
-                     os.path.join(home, '.nvpy.cfg'),
-                     os.path.join(home, '.nvpy'),
-                     os.path.join(home, '.nvpyrc')]
+        cfg_files = [
+            os.path.join(app_dir, 'nvpy.cfg'),
+            os.path.join(home, 'nvpy.cfg'),
+            os.path.join(home, '.nvpy.cfg'),
+            os.path.join(home, '.nvpy'),
+            os.path.join(home, '.nvpyrc')
+        ]
 
         # user has specified either a specific path to a CFG file, or a
         # path relative to the nvpy.py location. specific takes precedence.
         if args is not None and args.cfg is not None:
             cfg_files.extend([os.path.join(app_dir, args.cfg), args.cfg])
 
-        cp = ConfigParser.SafeConfigParser(defaults)
+        cp = ConfigParser(defaults)
         self.files_read = cp.read(cfg_files)
 
         cfg_sec = 'nvpy'
@@ -144,6 +189,7 @@ class Config:
         self.db_path = cp.get(cfg_sec, 'db_path')
         #  0 = alpha sort, 1 = last modified first
         self.notes_as_txt = cp.getint(cfg_sec, 'notes_as_txt')
+        self.read_txt_extensions = cp.get(cfg_sec, 'read_txt_extensions')
         self.txt_path = os.path.join(home, cp.get(cfg_sec, 'txt_path'))
         self.search_mode = cp.get(cfg_sec, 'search_mode')
         self.case_sensitive = cp.getint(cfg_sec, 'case_sensitive')
@@ -164,26 +210,49 @@ class Config:
         self.layout = cp.get(cfg_sec, 'layout')
         self.print_columns = cp.getint(cfg_sec, 'print_columns')
 
-        self.background_color = cp.get(cfg_sec, 'background_color')
+        self.colors = ColorConfig(
+            text=cp.get(cfg_sec, 'text_color'),
+            selected_note=cp.get(cfg_sec, 'selected_note_color'),
+            note_info=cp.get(cfg_sec, 'note_info_color'),
+            highlight_note_info=cp.get(cfg_sec, 'highlight_note_info_color'),
+            url=cp.get(cfg_sec, 'url_color'),
+            background=cp.get(cfg_sec, 'background_color'),
+            highlight_background=cp.get(cfg_sec, 'highlight_background_color'),
+        )
 
         self.rest_css_path = cp.get(cfg_sec, 'rest_css_path')
+        self.md_css_path = cp.get(cfg_sec, 'md_css_path')
+        self.md_extensions = cp.get(cfg_sec, 'md_extensions')
         self.debug = cp.getint(cfg_sec, 'debug')
+        self.keep_search_keyword = cp.getboolean(cfg_sec, 'keep_search_keyword')
+        self.confirm_delete = cp.getboolean(cfg_sec, 'confirm_delete')
+        self.confirm_exit = cp.getboolean(cfg_sec, 'confirm_exit')
+
+        self.warnings = []
+        if cp.has_option(cfg_sec, 'background_full_sync'):
+            w = lambda: logging.warning('"background_full_sync" option is removed.')
+            self.warnings.append(w)
 
     def parse_cmd_line_opts(self):
         if __name__ != '__main__':
             return None
 
         parser = argparse.ArgumentParser()
-        parser.add_argument('--cfg', '-c', default='', dest='cfg',
-                            metavar='nvpy.cfg', help='path to config file')
+        parser.add_argument('--cfg', '-c', default='', dest='cfg', metavar='nvpy.cfg', help='path to config file')
         args = parser.parse_args()
         return args
 
+    def show_warnings(self):
+        """ Show warnings when using obsoleted option. """
+        for w in self.warnings:
+            w()
+
 
 class NotesListModel(SubjectMixin):
     """
     @ivar list: List of (str key, dict note) objects.
     """
+
     def __init__(self):
         # call mixin ctor
         SubjectMixin.__init__(self)
@@ -205,12 +274,21 @@ class NotesListModel(SubjectMixin):
         else:
             return -1
 
+    def get(self, key):
+        idx = self.get_idx(key)
+        if idx < 0:
+            raise KeyError('Note is not found: key={}'.format(key))
+
+        return self.list[idx]
+
 
 class Controller:
     """Main application class.
     """
 
     def __init__(self, config):
+        SubjectMixin.MAIN_THREAD = threading.current_thread()
+
         # should probably also look in $HOME
         self.config = config
         self.config.app_version = VERSION
@@ -224,7 +302,7 @@ class Controller:
 
         log_filename = os.path.join(self.config.db_path, 'nvpy.log')
         # file will get nuked when it reaches 100kB
-        lhandler = RotatingFileHandler(log_filename, maxBytes=100000, backupCount=1)
+        lhandler = RotatingFileHandler(log_filename, maxBytes=100000, backupCount=1, encoding='utf8')
         lhandler.setLevel(logging.DEBUG)
         lhandler.setFormatter(logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(message)s'))
         # we get the root logger and configure it
@@ -235,93 +313,105 @@ class Controller:
         # this will go to the root logger
         logging.debug('nvpy logging initialized')
 
-        logging.debug('config read from %s' % (str(self.config.files_read),))
+        logging.debug('config read from %s' % (str(self.config.files_read), ))
+        self.config.show_warnings()
 
         if self.config.sn_username == '':
             self.config.simplenote_sync = 0
 
-        css = self.config.rest_css_path
-        if css:
-            if css.startswith("~/"):
+        rst_css = self.config.rest_css_path
+        if rst_css:
+            if rst_css.startswith("~/"):
                 # On Mac, paths that start with '~/' aren't found by path.exists
-                css = css.replace(
-                    "~", os.path.abspath(os.path.expanduser('~')), 1)
-                self.config.rest_css_path = css
-            if not os.path.exists(css):
+                rst_css = rst_css.replace("~", os.path.abspath(os.path.expanduser('~')), 1)
+                self.config.rest_css_path = rst_css
+            if not os.path.exists(rst_css):
                 # Couldn't find the user-defined css file. Use docutils css instead.
                 self.config.rest_css_path = None
+        md_css = self.config.md_css_path
+        if md_css:
+            if md_css.startswith("~/"):
+                # On Mac, paths that start with '~/' aren't found by path.exists
+                md_css = md_css.replace("~", os.path.abspath(os.path.expanduser('~')), 1)
+                self.config.md_css_path = md_css
+            if not os.path.exists(md_css):
+                # Couldn't find the user-defined css file.
+                # Do not use css styling for markdown.
+                self.config.md_css_path = None
 
         self.notes_list_model = NotesListModel()
         # create the interface
         self.view = view.View(self.config, self.notes_list_model)
 
-        # read our database of notes into memory
-        # and sync with simplenote.
         try:
-            self.notes_db = NotesDB(self.config)
+            # read our database of notes into memory
+            # and sync with simplenote.
+            try:
+                self.notes_db = NotesDB(self.config)
+            except ReadError as e:
+                emsg = "Please check nvpy.log.\n" + str(e)
+                self.view.show_error('Sync error', emsg)
+                exit(1)
 
-        except ReadError, e:
-            emsg = "Please check nvpy.log.\n" + str(e)
-            self.view.show_error('Sync error', emsg)
-            exit(1)
+            self.notes_db.add_observer('synced:note', self.observer_notes_db_synced_note)
+            self.notes_db.add_observer('change:note-status', self.observer_notes_db_change_note_status)
 
-        self.notes_db.add_observer('synced:note', self.observer_notes_db_synced_note)
-        self.notes_db.add_observer('change:note-status', self.observer_notes_db_change_note_status)
+            if self.config.simplenote_sync:
+                self.notes_db.add_observer('progress:sync_full', self.observer_notes_db_sync_full)
+                self.notes_db.add_observer('error:sync_full', self.observer_notes_db_error_sync_full)
+                self.notes_db.add_observer('complete:sync_full', self.observer_notes_db_complete_sync_full)
 
-        if self.config.simplenote_sync:
-            self.notes_db.add_observer('progress:sync_full', self.observer_notes_db_sync_full)
-            self.sync_full()
+            # we want to be notified when the user does stuff
+            self.view.add_observer('click:notelink', self.observer_view_click_notelink)
+            self.view.add_observer('delete:note', self.observer_view_delete_note)
+            self.view.add_observer('select:note', self.observer_view_select_note)
+            self.view.add_observer('change:entry', self.observer_view_change_entry)
+            self.view.add_observer('change:text', self.observer_view_change_text)
+            self.view.add_observer('change:pinned', self.observer_view_change_pinned)
+            self.view.add_observer('create:note', self.observer_view_create_note)
+            self.view.add_observer('keep:house', self.observer_view_keep_house)
+            self.view.add_observer('command:markdown', self.observer_view_markdown)
+            self.view.add_observer('command:rest', self.observer_view_rest)
+            self.view.add_observer('delete:tag', self.observer_view_delete_tag)
+            self.view.add_observer('add:tag', self.observer_view_add_tag)
 
-        # we want to be notified when the user does stuff
-        self.view.add_observer('click:notelink',
-                self.observer_view_click_notelink)
-        self.view.add_observer('delete:note', self.observer_view_delete_note)
-        self.view.add_observer('select:note', self.observer_view_select_note)
-        self.view.add_observer('change:entry', self.observer_view_change_entry)
-        self.view.add_observer('change:text', self.observer_view_change_text)
-        self.view.add_observer('change:pinned', self.observer_view_change_pinned)
-        self.view.add_observer('create:note', self.observer_view_create_note)
-        self.view.add_observer('keep:house', self.observer_view_keep_house)
-        self.view.add_observer('command:markdown', self.observer_view_markdown)
-        self.view.add_observer('command:rest', self.observer_view_rest)
-        self.view.add_observer('delete:tag', self.observer_view_delete_tag)
-        self.view.add_observer('add:tag', self.observer_view_add_tag)
+            if self.config.simplenote_sync:
+                self.view.add_observer('command:sync_full', lambda v, et, e: self.sync_full())
+                self.view.add_observer('command:sync_current_note', self.observer_view_sync_current_note)
 
-        if self.config.simplenote_sync:
-            self.view.add_observer('command:sync_full', lambda v, et, e: self.sync_full())
-            self.view.add_observer('command:sync_current_note', self.observer_view_sync_current_note)
+            self.view.add_observer('close', self.observer_view_close)
 
-        self.view.add_observer('close', self.observer_view_close)
+            # setup UI to reflect our search mode and case sensitivity
+            self.view.set_cs(self.config.case_sensitive, silent=True)
+            self.view.set_search_mode(self.config.search_mode, silent=True)
 
-        # setup UI to reflect our search mode and case sensitivity
-        self.view.set_cs(self.config.case_sensitive, silent=True)
-        self.view.set_search_mode(self.config.search_mode, silent=True)
+            self.view.add_observer('change:cs', self.observer_view_change_cs)
+            self.view.add_observer('change:search_mode', self.observer_view_change_search_mode)
 
-        self.view.add_observer('change:cs', self.observer_view_change_cs)
-        self.view.add_observer('change:search_mode', self.observer_view_change_search_mode)
+            # nn is a list of (key, note) objects
+            nn, match_regexp, active_notes = self.notes_db.filter_notes()
+            # this will trigger the list_change event
+            self.notes_list_model.set_list(nn)
+            self.notes_list_model.match_regexp = match_regexp
+            self.view.set_note_tally(len(nn), active_notes, len(self.notes_db.notes))
 
-        # nn is a list of (key, note) objects
-        nn, match_regexp, active_notes = self.notes_db.filter_notes()
-        # this will trigger the list_change event
-        self.notes_list_model.set_list(nn)
-        self.notes_list_model.match_regexp = match_regexp
-        self.view.set_note_tally(len(nn), active_notes, len(self.notes_db.notes))
+            # we'll use this to keep track of the currently selected note
+            # we only use idx, because key could change from right under us.
+            self.selected_note_key = None
+            self.view.select_note(0)
 
-        # we'll use this to keep track of the currently selected note
-        # we only use idx, because key could change from right under us.
-        self.selected_note_idx = -1
-        self.view.select_note(0)
-
-    def get_selected_note_key(self):
-        if self.selected_note_idx >= 0:
-            return self.notes_list_model.list[self.selected_note_idx].key
-        else:
-            return None
+            if self.config.simplenote_sync:
+                self.view.after(0, self.sync_full)
+        except BaseException:
+            # Initialization failed.  Stop all timers.
+            self.view.cancel_timers()
+            raise
 
     def main_loop(self):
         if not self.config.files_read:
-            self.view.show_warning('No config file',
-                                  'Could not read any configuration files. See https://github.com/cpbotha/nvpy for details.')
+            self.view.show_warning(
+                'No config file',
+                'Could not read any configuration files. See https://github.com/cpbotha/nvpy for details.')
 
         elif not self.config.ok:
             wmsg = ('Please rename [default] to [nvpy] in %s. ' + \
@@ -329,35 +419,83 @@ class Controller:
             (str(self.config.files_read),)
             self.view.show_warning('Rename config section', wmsg)
 
-        self.view.main_loop()
+        def poll_notifies():
+            self.view.after(100, poll_notifies)
+            self.notes_db.handle_notifies()
 
-    def observer_notes_db_change_note_status(self, notes_db, evt_type, evt):
-        skey = self.get_selected_note_key()
+        self.view.after(0, poll_notifies)
+        try:
+            self.view.main_loop()
+        finally:
+            # Cancel all timers before stop this program.
+            self.view.cancel_timers()
+
+    def observer_notes_db_change_note_status(self, notes_db, evt_type, evt: events.NoteStatusChangedEvent):
+        skey = self.selected_note_key
         if skey == evt.key:
             self.view.set_note_status(self.notes_db.get_note_status(skey))
 
-    def observer_notes_db_sync_full(self, notes_db, evt_type, evt):
+    def observer_notes_db_sync_full(self, notes_db, evt_type, evt: events.SyncProgressEvent):
         logging.debug(evt.msg)
         self.view.set_status_text(evt.msg)
 
-    def observer_notes_db_synced_note(self, notes_db, evt_type, evt):
+        # regenerate display list
+        # reselect old selection
+        # put cursor where it used to be.
+        self.view.refresh_notes_list()
+
+        # change status to "Full syncing"
+        self.update_note_status()
+
+    def observer_notes_db_error_sync_full(self, notes_db, evt_type, evt: events.SyncFailedEvent):
+        try:
+            raise evt.error
+        except SyncError as e:
+            self.view.show_error('Sync error', e)
+        except WriteError as e:
+            emsg = "Please check nvpy.log.\n" + str(e)
+            self.view.show_error('Sync error', emsg)
+            exit(1)
+        except Exception as e:
+            crash_log = ''.join(traceback.format_exception(*evt.exc_info))
+            logging.error(crash_log)
+            emsg = 'An unexpected error has occurred.\n'\
+                   'Please check nvpy.log.\n' \
+                   + repr(e)
+            self.view.show_error('Sync error', emsg)
+            exit(1)
+
+        # return normal status from "Full syning".
+        self.update_note_status()
+
+    def observer_notes_db_complete_sync_full(self, notes_db, evt_type, evt: events.SyncCompletedEvent):
+        sync_from_server_errors = evt.errors
+        if sync_from_server_errors > 0:
+            self.view.show_error(
+                'Error syncing notes from server',
+                'Error syncing %d notes from server. Please check nvpy.log for details.' % (sync_from_server_errors, ))
+
+        # return normal status from "Full syning".
+        self.update_note_status()
+
+    def observer_notes_db_synced_note(self, notes_db, evt_type, evt: events.NoteSyncedEvent):
         """This observer gets called only when a note returns from
         a sync that's more recent than our most recent mod to that note.
         """
 
-        selected_note_o = self.notes_list_model.list[self.selected_note_idx]
         # if the note synced back matches our currently selected note,
         # we overwrite.
-
-        if selected_note_o.key == evt.lkey:
+        if self.selected_note_key is not None and self.selected_note_key == evt.lkey:
+            selected_note_o = self.notes_list_model.get(self.selected_note_key)
             if selected_note_o.note['content'] != evt.old_note['content']:
                 self.view.mute_note_data_changes()
                 # in this case, we want to keep the user's undo buffer so that they
                 # can undo synced back changes if they would want to.
                 self.view.set_note_data(selected_note_o.note, reset_undo=False)
                 self.view.unmute_note_data_changes()
+        self.view.refresh_notes_list()
 
-    def observer_view_click_notelink(self, view, evt_type, note_name):
+    def observer_view_click_notelink(self, view, evt_type, note_name: str):
         # find note_name in titles, try to jump to that note
         # if not in current list, change search string in case
         # it's somewhere else
@@ -370,19 +508,16 @@ class Controller:
             # to be a new regular expression to search for in the notes db.
             self.view.set_search_entry_text(note_name)
 
-    def observer_view_delete_note(self, view, evt_type, evt):
+    def observer_view_delete_note(self, view, evt_type, evt: events.NoteSelectionChangedEvent):
         # delete note from notes_db
         # remove the note from the notes_list_model.list
 
-        # if these two are not equal, something is not kosher.
-        assert(evt.sel == self.selected_note_idx)
-
         # first get key of note that is to be deleted
-        key = self.get_selected_note_key()
+        key = self.selected_note_key
 
         # then try to select after the one that is to be deleted
         nidx = evt.sel + 1
-        if nidx >= 0 and nidx < self.view.get_number_of_notes():
+        if 0 <= nidx < self.view.get_number_of_notes():
             self.view.select_note(nidx)
 
         # finally delete the note
@@ -394,14 +529,24 @@ class Controller:
         self.view.set_search_entry_text(self.view.get_search_entry_text())
 
     def helper_markdown_to_html(self):
-        if self.selected_note_idx >= 0:
-            key = self.notes_list_model.list[self.selected_note_idx].key
+        if self.selected_note_key:
+            key = self.selected_note_key
             c = self.notes_db.get_note_content(key)
-            logging.debug("Trying to convert %s to html." % (key,))
+            logging.debug("Trying to convert %s to html." % (key, ))
             if HAVE_MARKDOWN:
-                logging.debug("Convert note %s to html." % (key,))
-                html = markdown.markdown(c)
+                logging.debug("Convert note %s to html." % (key, ))
+                exts = re.split("\\s", self.config.md_extensions.strip()) if self.config.md_extensions else []
+                exts += list(DEFAULT_MARKDOWN_EXTS)
+                # remove duplicate items on exts.
+                exts = list(set(exts))
+
+                html = markdown.markdown(c, extensions=exts)
                 logging.debug("Convert done.")
+                if self.config.md_css_path:
+                    css = u"""<link rel="stylesheet" href="%s">""" % (self.config.md_css_path, )
+                    html = u"""<div class="markdown-body">%s</div>""" % (html, )
+                else:
+                    css = u""""""
 
             else:
                 logging.debug("Markdown not installed.")
@@ -416,27 +561,31 @@ class Controller:
 <head>
 <meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
 %s
+%s
 </head>
 <body>
 %s
 </body>
 </html>
-            """ % ('<meta http-equiv="refresh" content="5">' if self.view.get_continuous_rendering() else "",html,)
+            """ % (
+                '<meta http-equiv="refresh" content="5">' if self.view.get_continuous_rendering() else "",
+                css if self.config.md_css_path else "",
+                html,
+            )
             f.write(s)
             f.close()
             return fn
 
     def helper_rest_to_html(self):
-        if self.selected_note_idx >= 0:
-            key = self.notes_list_model.list[self.selected_note_idx].key
+        if self.selected_note_key:
+            key = self.selected_note_key
             c = self.notes_db.get_note_content(key)
             if HAVE_DOCUTILS:
                 settings = {}
                 if self.config.rest_css_path:
                     settings['stylesheet_path'] = self.config.rest_css_path
                 # this gives the whole document
-                html = docutils.core.publish_string(
-                    c, writer_name='html', settings_overrides=settings)
+                html = docutils.core.publish_string(c, writer_name='html', settings_overrides=settings)
                 # publish_parts("*anurag*",writer_name='html')['body']
                 # gives just the desired part of the tree
 
@@ -453,7 +602,7 @@ class Controller:
             # "'ascii' codec can't decode byte" error
             s = u"""
 %s
-            """ % (unicode(html, 'utf8'),)
+            """ % (unicode(html, 'utf8'), )
 
             f.write(s)
             f.close()
@@ -484,8 +633,8 @@ class Controller:
         else:
             syncn = wfsn = 0
 
-        savet = 'Saving %d notes.' % (saven,) if saven > 0 else ''
-        synct = 'Waiting to sync %d notes.' % (syncn,) if syncn > 0 else ''
+        savet = 'Saving %d notes.' % (saven, ) if saven > 0 else ''
+        synct = 'Waiting to sync %d notes.' % (syncn, ) if syncn > 0 else ''
         wfsnt = 'Syncing with simplenote server.' if wfsn else ''
 
         return ' '.join([i for i in [savet, synct, wfsnt] if i])
@@ -507,43 +656,39 @@ class Controller:
         if self.view.get_continuous_rendering():
             self.helper_markdown_to_html()
 
-    def observer_view_select_note(self, view, evt_type, evt):
+    def observer_view_select_note(self, view, evt_type, evt: events.NoteSelectionChangedEvent):
         self.select_note(evt.sel)
 
     def observer_view_sync_current_note(self, view, evt_type, evt):
-        if self.selected_note_idx >= 0:
-            key = self.notes_list_model.list[self.selected_note_idx].key
+        if self.selected_note_key:
+            key = self.selected_note_key
             # this call will update our in-memory version if necessary
             ret = self.notes_db.sync_note_unthreaded(key)
             if ret and ret[1] == True:
-                self.view.update_selected_note_data(
-                        self.notes_db.notes[key])
-                self.view.set_status_text(
-                'Synced updated note from server.')
+                self.view.update_selected_note_data(self.notes_db.notes[key])
+                self.view.set_status_text('Synced updated note from server.')
 
             elif ret[1] == False:
-                self.view.set_status_text(
-                        'Server had nothing newer for this note.')
+                self.view.set_status_text('Server had nothing newer for this note.')
 
             elif ret is None:
-                self.view.set_status_text(
-                        'Unable to sync with server. Offline?')
+                self.view.set_status_text('Unable to sync with server. Offline?')
 
-    def observer_view_change_cs(self, view, evt_type, evt):
+    def observer_view_change_cs(self, view, evt_type, evt: events.CheckboxChangedEvent):
         # evt.value is the new value
         # only do something if user has really toggled
         if evt.value != self.config.case_sensitive:
             self.config.case_sensitive = evt.value
             self.view.refresh_notes_list()
 
-    def observer_view_change_search_mode(self, view, evt_type, evt):
+    def observer_view_change_search_mode(self, view, evt_type, evt: events.CheckboxChangedEvent):
         if evt.value != self.config.search_mode:
             self.config.search_mode = evt.value
             self.view.refresh_notes_list()
 
-    def observer_view_change_entry(self, view, evt_type, evt):
+    def observer_view_change_entry(self, view, evt_type, evt: events.TextBoxChangedEvent):
         # store the currently selected note key
-        k = self.get_selected_note_key()
+        k = self.selected_note_key
         # for each new evt.value coming in, get a new list from the notes_db
         # and set it in the notes_list_model
         nn, match_regexp, active_notes = self.notes_db.filter_notes(evt.value)
@@ -567,8 +712,6 @@ class Controller:
             # if it does turn out to be new note content, this will be handled
             # a few lines down.
             self.view.select_note(idx, silent=True)
-            # but of course we DO have to record the possibly new IDX!!
-            self.selected_note_idx = idx
 
             # see if the note has been updated (content, tags, pin)
             new_note = self.notes_db.get_note(k)
@@ -577,7 +720,7 @@ class Controller:
             # currently being displayed. this could happen if a sync gets
             # a new note of the server to replace the currently displayed one.
             if self.view.is_note_different(new_note):
-                logging.debug("Currently selected note %s replaced by newer from server." % (k,))
+                logging.debug("Currently selected note %s replaced by newer from server." % (k, ))
                 # carefully update currently selected note
                 # restore cursor position, search and link highlights
                 self.view.update_selected_note_data(new_note)
@@ -591,36 +734,28 @@ class Controller:
     def observer_view_change_text(self, view, evt_type, evt):
         # get new text and update our database
         # need local key of currently selected note for this
-        if self.selected_note_idx >= 0:
-            key = self.notes_list_model.list[self.selected_note_idx].key
-            self.notes_db.set_note_content(key,
-                                           self.view.get_text())
+        if self.selected_note_key:
+            self.notes_db.set_note_content(self.selected_note_key, self.view.get_text())
 
     def observer_view_change_tags(self, view, evt_type, evt):
         # get new text and update our database
-        # need local key of currently selected note for this
-        if self.selected_note_idx >= 0:
-            key = self.notes_list_model.list[self.selected_note_idx].key
-            self.notes_db.set_note_tags(key, evt.value)
+        if self.selected_note_key:
+            self.notes_db.set_note_tags(self.selected_note_key, evt.value)
             self.view.cmd_notes_list_select()
 
-    def observer_view_delete_tag(self, view, evt_type, evt):
-        key = self.notes_list_model.list[self.selected_note_idx].key
-        self.notes_db.delete_note_tag(key, evt.tag)
+    def observer_view_delete_tag(self, view, evt_type, evt: events.TagRemovedEvent):
+        self.notes_db.delete_note_tag(self.selected_note_key, evt.tag)
         self.view.cmd_notes_list_select()
-    
-    def observer_view_add_tag(self, view, evt_type, evt):
-        key = self.notes_list_model.list[self.selected_note_idx].key
-        self.notes_db.add_note_tags(key, evt.tags)
+
+    def observer_view_add_tag(self, view, evt_type, evt: events.TagsAddedEvent):
+        self.notes_db.add_note_tags(self.selected_note_key, evt.tags)
         self.view.cmd_notes_list_select()
         self.view.tags_entry_var.set('')
 
-    def observer_view_change_pinned(self, view, evt_type, evt):
+    def observer_view_change_pinned(self, view, evt_type, evt: events.CheckboxChangedEvent):
         # get new text and update our database
-        # need local key of currently selected note for this
-        if self.selected_note_idx >= 0:
-            key = self.notes_list_model.list[self.selected_note_idx].key
-            self.notes_db.set_note_pinned(key, evt.value)
+        if self.selected_note_key:
+            self.notes_db.set_note_pinned(self.selected_note_key, evt.value)
 
     def observer_view_close(self, view, evt_type, evt):
         # check that everything has been saved and synced before exiting
@@ -646,13 +781,21 @@ class Controller:
                 self.view.close()
 
         else:
+            if self.config.confirm_exit:
+                msg = "Do you want to exit?"
+                if not self.view.askyesno('Confirm exit', msg):
+                    return
+
             self.view.close()
 
-    def observer_view_create_note(self, view, evt_type, evt):
+    def observer_view_create_note(self, view, evt_type, evt: events.NoteCreatedEvent):
         # create the note
         new_key = self.notes_db.create_note(evt.title)
         # clear the search entry, this should trigger a new list being returned
-        self.view.set_search_entry_text('')
+        keyword = ''
+        if self.config.keep_search_keyword:
+            keyword = self.view.get_search_entry_text()
+        self.view.set_search_entry_text(keyword)
         # we should focus on our thingy
         idx = self.notes_list_model.get_idx(new_key)
         self.view.select_note(idx)
@@ -675,14 +818,13 @@ class Controller:
         else:
             key = None
             note = None
-            idx = -1
             # no note selected, so we clear the UI (and display a clear
             # message that no note is selected) and we disable note
             # editing controls.
             self.view.clear_note_ui()
             self.view.set_note_editing(False)
 
-        self.selected_note_idx = idx
+        self.selected_note_key = key
 
         # when we do this, we don't want the change:{text,tags,pinned} events
         # because those should only fire when they are changed through the UI
@@ -694,30 +836,32 @@ class Controller:
         self.view.unmute_note_data_changes()
 
     def sync_full(self):
-        try:
-            sync_from_server_errors = self.notes_db.sync_full()
+        self.notes_db.sync_full_threaded()
 
-        except SyncError, e:
-            self.view.show_error('Sync error', e)
-        except WriteError, e:
-            emsg = "Please check nvpy.log.\n" + str(e)
-            self.view.show_error('Sync error', emsg)
-            exit(1)
-
-        else:
-            # regenerate display list
-            # reselect old selection
-            # put cursor where it used to be.
-            self.view.refresh_notes_list()
-
-            if sync_from_server_errors > 0:
-                self.view.show_error('Error syncing notes from server', 'Error syncing %d notes from server. Please check nvpy.log for details.' % (sync_from_server_errors,))
+    def update_note_status(self):
+        skey = self.selected_note_key
+        self.view.set_note_status(self.notes_db.get_note_status(skey))
 
 
 def main():
     # setup appdir
     if hasattr(sys, 'frozen') and sys.frozen:
-        appdir, _ = os.path.split(sys.executable)
+        if hasattr(sys, '_MEIPASS'):
+            # PyInstaller
+            appdir = sys._MEIPASS
+
+            # WORKAROUND: Bug that always raise the SSLCertVerificationError from urlopen()
+            #             when CPython is not installed.
+
+            # Use certificate from certifi only if cafile could not find by ssl.
+            # See https://github.com/pyinstaller/pyinstaller/pull/3952
+            import ssl
+            if ssl.get_default_verify_paths().cafile is None:
+                import certifi
+                os.environ['SSL_CERT_FILE'] = certifi.core.where()
+        else:
+            # py2exe
+            appdir, _ = os.path.split(sys.executable)
 
     else:
         dirname, _ = os.path.split(os.path.realpath(__file__))
@@ -731,8 +875,14 @@ def main():
 
     config = Config(appdir_full_path)
 
-    controller = Controller(config)
-    controller.main_loop()
+    try:
+        controller = Controller(config)
+        controller.main_loop()
+    except tk.Ucs4NotSupportedError as e:
+        logging.error(str(e))
+        import tkMessageBox
+        tkMessageBox.showerror('UCS-4 not supported', str(e))
+        raise
 
 
 if __name__ == '__main__':
