@@ -1,12 +1,16 @@
 # nvPY: cross-platform note-taking app with simplenote syncing
 # copyright 2012 by Charl P. Botha <cpbotha@vxlabs.com>
 # new BSD license
-
 import datetime
 import random
 import re
-import string
-import urllib2
+from urllib.request import urlopen, URLError
+from queue import Queue, Empty as QueueEmpty
+from .p3port import unicode
+
+import threading
+
+from . import tk
 
 # first line with non-whitespace should be the title
 note_title_re = re.compile('\s*(.*)\n?')
@@ -17,7 +21,7 @@ def generate_random_key():
 
     stackoverflow question 2782229
     """
-    return '%030x' % (random.randrange(256 ** 15),)
+    return '%030x' % (random.randrange(256**15), )
 
 
 def get_note_title(note):
@@ -104,6 +108,7 @@ def note_markdown(n):
     else:
         return 0
 
+
 tags_illegal_chars = re.compile(r'[\s]')
 
 
@@ -130,31 +135,23 @@ def sanitise_tags(tags):
         return illegals_removed.split(',')
 
 
-def sort_by_title_pinned(a, b):
-    if note_pinned(a.note) and not note_pinned(b.note):
-        return -1
-    elif not note_pinned(a.note) and note_pinned(b.note):
-        return 1
-    else:
-        return cmp(get_note_title(a.note), get_note_title(b.note))
+def sort_key_by_title_pinned(a):
+    if note_pinned(a.note):
+        return (1, get_note_title(a.note))
+    return (0, get_note_title(a.note))
 
 
-def sort_by_modify_date_pinned(a, b):
-    if note_pinned(a.note) and not note_pinned(b.note):
-        return 1
-    elif not note_pinned(a.note) and note_pinned(b.note):
-        return -1
-    else:
-        return cmp(float(a.note.get('modifydate', 0)), float(b.note.get('modifydate', 0)))
+def sort_key_by_modify_date_pinned(a):
+    if note_pinned(a.note):
+        return (1, float(a.note.get('modifydate', 0)))
+    return (0, float(a.note.get('modifydate', 0)))
 
 
-def sort_by_create_date_pinned(a, b):
-    if note_pinned(a.note) and not note_pinned(b.note):
-        return 1
-    elif not note_pinned(a.note) and note_pinned(b.note):
-        return -1
-    else:
-        return cmp(float(a.note.get('createdate', 0)), float(b.note.get('createdate', 0)))
+def sort_key_by_create_date_pinned(a):
+    if note_pinned(a.note):
+        return (1, float(a.note.get('createdate', 0)))
+    return (0, float(a.note.get('createdate', 0)))
+
 
 def check_internet_on():
     """Utility method to check if we have an internet connection.
@@ -162,36 +159,37 @@ def check_internet_on():
     slightly adapted from: http://stackoverflow.com/a/3764660/532513
     """
     try:
-        urllib2.urlopen('http://74.125.228.100', timeout=1)
+        urlopen('http://74.125.228.100', timeout=1)
         return True
 
-    except urllib2.URLError:
+    except URLError:
         pass
 
     return False
 
 
-class KeyValueObject:
-    """Store key=value pairs in this object and retrieve with o.key.
-
-    You should also be able to do MiscObject(**your_dict) for the same effect.
-    """
-
-    def __init__(self, **kwargs):
-        self.__dict__.update(kwargs)
-
-
 class SubjectMixin:
     """Maintain a list of callables for each event type.
 
     We follow the convention action:object, e.g. change:entry.
     """
 
+    # main thread object
+    #
+    # Workaround for missing the threading.main_thread() function on Python 2.7.
+    # This variable was updated by the Controller.__init__().
+    MAIN_THREAD = None
+
     def __init__(self):
         self.observers = {}
         self.mutes = {}
+        self.notifies = Queue()
 
     def add_observer(self, evt_type, o):
+        from .debug import wrap_buggy_function
+        o = tk.with_ucs4_error_handling(o)
+        o = wrap_buggy_function(o)
+
         if evt_type not in self.observers:
             self.observers[evt_type] = [o]
 
@@ -202,9 +200,33 @@ class SubjectMixin:
         if evt_type in self.mutes or evt_type not in self.observers:
             return
 
-        for o in self.observers[evt_type]:
-            # invoke observers with ourselves as first param
-            o(self, evt_type, evt)
+        if threading.current_thread() == self.MAIN_THREAD:
+            for o in self.observers[evt_type]:
+                self.__invoke_observer(o, evt_type, evt)
+
+        else:
+            # Tkinter is not thread safe. so, observers must be executed on MAIN_THREAD.
+            self.notifies.put((evt_type, evt))
+
+    def handle_notifies(self):
+        """Do processing to queued notifies.
+        
+        This method is periodically called from main thread.
+        """
+        assert threading.current_thread() == self.MAIN_THREAD
+        try:
+            while True:
+                evt_type, evt = self.notifies.get_nowait()
+
+                for o in self.observers[evt_type]:
+                    self.__invoke_observer(o, evt_type, evt)
+
+        except QueueEmpty:
+            pass
+
+    def __invoke_observer(self, observer, event_type, event):
+        # invoke observers with ourselves as first param
+        observer(self, event_type, event)
 
     def mute(self, evt_type):
         self.mutes[evt_type] = True
